-------------------------------------------------------------------------------
-- $Id: defer_state.vhd,v 1.1 2007/12/11 06:41:02 sanjayk Exp $
-------------------------------------------------------------------------------
-- defer_state - entity/architecture pair
-------------------------------------------------------------------------------
--  ***************************************************************************
--  **  Copyright(C) 2006 by Xilinx, Inc. All rights reserved.               **
--  **                                                                       **
--  **  This text contains proprietary, confidential                         **
--  **  information of Xilinx, Inc. , is distributed by                      **
--  **  under license from Xilinx, Inc., and may be used,                    **
--  **  copied and/or disclosed only pursuant to the terms                   **
--  **  of a valid license agreement with Xilinx, Inc.                       **
--  **                                                                       **
--  **  Unmodified source code is guaranteed to place and route,             **
--  **  function and run at speed according to the datasheet                 **
--  **  specification. Source code is provided "as-is", with no              **
--  **  obligation on the part of Xilinx to provide support.                 **
--  **                                                                       **
--  **  Xilinx Hotline support of source code IP shall only include          **
--  **  standard level Xilinx Hotline support, and will only address         **
--  **  issues and questions related to the standard released Netlist        **
--  **  version of the core (and thus indirectly, the original core source). **
--  **                                                                       **
--  **  The Xilinx Support Hotline does not have access to source            **
--  **  code and therefore cannot answer specific questions related          **
--  **  to source HDL. The Xilinx Support Hotline will only be able          **
--  **  to confirm the problem in the Netlist version of the core.           **
--  **                                                                       **
--  **  This copyright and support notice must be retained as part           **
--  **  of this text at all times.                                           **
--  ***************************************************************************
--
-------------------------------------------------------------------------------
-- Filename:        defer_state.vhd
-- 
-- Description:     
--                  This file contains the transmit deferral state machine.
-- VHDL-Standard:   VHDL'93
-------------------------------------------------------------------------------
-- Structure:   
--
--  xps_ethernetlite.vhd
--      \
--      \-- xemac.vhd
--           \
--           \--xps_ipif_ssp1.vhd
--           \
--           \-- emac_dpram.vhd                     
--           \    \                     
--           \    \-- RAMB16_S36_S36
--           \    \     or
--           \    \-- RAMB4_S8_S8
--           \                          
--           \-- emac.vhd                     
--                \                     
--                \                     
--                \-- receive.vhd
--                \      rx_state_2.vhd
--                \      rx_intrfce.vhd
--                \         veasyncfifo5w15ddstrb.edn
--                \            or
--                \         v2asyncfifo5w15ddstrb.edn
--                \      cntr4bit.vhd
--                \      nibshiftin.vhd
--                \      nibshiftdain.vhd
--                \      cntr11bit.vhd
--                \      cntr12bit.vhd
--                \      dacheck.vhd
--                \      rxlengthcntr.vhd
--                \      crcgenrx.vhd
--                \      rxcrcnibcnt.vhd
--                \                     
--                \-- transmit_2.vhd
--                       crcgentx.vhd
--                          crcnibshiftreg
--                       tx_intrfce.vhd
--                          veasyncfifo5w15ddstrb.edn
--                             or
--                          v2asyncfifo5w15ddstrb.edn
--                       cntr11bit.vhd
--                       nibshiftout.vhd
--                       tx_state_3.vhd
--                       deferral.vhd
--                          cntr5bit.vhd
--                          defer_state.vhd
--                       bocntr.vhd
--                          lfsr16.vhd
--                       msh_cnt.vhd
--                       ld_arith_reg.vhd
--
-------------------------------------------------------------------------------
-- Author:         MSH
-- History:
--  RSK            11/06/06
-- ^^^^^^
--                 First Version, Based on opb_ethernetlite v1.01.b
--  ~~~~~
-------------------------------------------------------------------------------
-- Naming Conventions:
--      active low signals:                     "*_n"
--      clock signals:                          "clk", "clk_div#", "clk_#x" 
--      reset signals:                          "rst", "rst_n" 
--      generics:                               "C_*" 
--      user defined types:                     "*_TYPE" 
--      state machine next state:               "*_ns" 
--      state machine current state:            "*_cs" 
--      combinatorial signals:                  "*_com" 
--      pipelined or register delay signals:    "*_d#" 
--      counter signals:                        "*cnt*"
--      clock enable signals:                   "*_ce" 
--      internal version of output port         "*_i"
--      device pins:                            "*_pin" 
--      ports:                                  - Names begin with Uppercase 
--      processes:                              "*_PROCESS" 
--      component instantiations:               "<ENTITY_>I_<#|FUNC>
-------------------------------------------------------------------------------
-- 
library ieee;
use ieee.STD_LOGIC_1164.all;

--Package file that contains constant definition for RESET_ACTIVE
library xps_ethernetlite_v2_00_c;
use xps_ethernetlite_v2_00_c.mac_pkg.all;

-- synopsys translate_off
Library XilinxCoreLib;
-- synopsys translate_on

-------------------------------------------------------------------------------
-- Port Declaration
-------------------------------------------------------------------------------

entity defer_state is
  port (
        txEn        : in std_logic; 
        txrst       : in std_logic;
        ifgp2Done   : in std_logic;
        ifgp1Done   : in std_logic;
        backingOff  : in std_logic;
        crs         : in std_logic;
        full_half_n : in std_logic;        
        deferring   : out std_logic;
        cntrEnbl    : out std_logic;
        cntrLd      : out std_logic;
        clk         : in std_logic;
        rst         : in std_logic
        );
 
end defer_state;

-------------------------------------------------------------------------------
-- Definition of Generics:
--          No Generics were used for this Entity.
--
-- Definition of Ports:
--         
-------------------------------------------------------------------------------

architecture implementation of defer_state is
-------------------------------------------------------------------------------
--  Constant Declarations
-------------------------------------------------------------------------------
-- Constants used in this design are found in mac_pkg.vhd
-------------------------------------------------------------------------------
-- Signal and Type Declarations
-------------------------------------------------------------------------------
 
  type StateName is (loadCntr,startIfgp1Cnt,startIfgp2Cnt,cntDone);
  signal thisState               : StateName;
  signal nextState               : StateName;  
   
-------------------------------------------------------------------------------
-- Component Declarations
-------------------------------------------------------------------------------
-- The following components are the building blocks of the tx state machine
 
begin

-------------------------------------------------------------------------------
-- An FSM that deals with transmitting data
-------------------------------------------------------------------------------
  
  FSMR : process (clk)
  begin  --
    if (clk'event and clk = '1') then     -- rising clock edge
      if (rst = '1' or txRst = '1') then
        thisState <= loadCntr;
      else
        thisState <= nextState;
      end if;
    end if;
  end process FSMR;

  FSMC : process (thisState,txEn,ifgp2Done,ifgp1Done,backingOff,crs,
                  full_half_n)
  begin  --
    case thisState is
      when loadCntr =>
        if (((txEn = '0') and (full_half_n = '1')) or
            ((crs = '0') and (full_half_n = '0') and (backingOff = '0'))) and 
            ifgp1Done = '0'  and ifgp2Done = '0' then
          nextState <= startIfgp1Cnt;
        else
          nextState <= loadCntr; -- wait for end of transmission
        end if;
      when startIfgp1Cnt =>
        if (((crs = '1') and (full_half_n = '0')) or
            ((backingOff = '1') and (full_half_n = '0'))) then
          nextState <= loadCntr;
        elsif (ifgp1Done = '1') then  -- gap done
          nextState <= startIfgp2Cnt;
        else
          nextState <= startIfgp1Cnt; -- still counting
        end if;
      when startIfgp2Cnt =>
        if (ifgp2Done = '1') then  -- gap done
          nextState <= cntDone;
        else
          nextState <= startIfgp2Cnt; -- still counting
        end if;        
      when cntDone =>
        if (txEn = '1' or crs = '1') then  -- transmission started
          nextState <= loadCntr;
        else
          nextState <= cntDone;
        end if;        
      when others         => null;
        nextState <= loadCntr;
    end case;
  end process FSMC;

  FSMD : process(thisState)
  
  begin  --
         
   if ((thisState =  loadCntr) or (thisState =  startIfgp1Cnt) or 
       (thisState =  startIfgp2Cnt)) then
     deferring <= '1';
   else
     deferring <= '0';
   end if;
   
   if ((thisState = startIfgp1Cnt) or (thisState =  startIfgp2Cnt)) then
     cntrEnbl   <= '1';
   else
     cntrEnbl   <= '0';
   end if;

   if (thisState = loadCntr) then
     cntrLd   <= '1';
   else
     cntrLd   <= '0';
   end if;
   
  end process FSMD;   

end implementation;

