/*! \file csmaMac.c
 \brief Carrier-Sensing Random Access MAC.
 
 @version 13
 @author Chris Hunter and Patrick Murphy
 

 The csmaMac is a modified ALOHA MAC that
 serves as an example for novel MAC
 development. Nodes transmit whenever
 they have information to transmit, and only
 move on to the next packet once the original
 transmit is acknowledged (ACKed). If no ACK
 is received, a collision is inferred and the
 packet is re-transmitted.
 
 
 By default, the MAC also implements carrier-
 sensing multiple-access with collision-
 avoidance (CSMA-CA). This functionality is
 built into hardware peripherals in the project
 so very little software state is affected.
 
 In its current state, the project acts as
 a true ethernet MAC-level wireless bridge.
 Any ethernet activity that appears on one
 WARP will be sent to another via the custom
 wireless link.
 
 Also, the current versions supports hardware-triggered
 ACK transmissions. This reduces the turn-around time
 as well as reduces the jitter on the start time of the ACK
 relative to the start time of the preceeding DATA.

 */
#include "warpnet_node.h"
#include "xtime_l.h"
#include "warpmac.h"
#include "warpphy.h"
#include "csmaMac.h"
#include "xparameters.h"
#include "string.h"
#include "errno.h"
#include "stdlib.h"
#include "stdio.h"
#include "ascii_characters.h"
#include "warp_userioboard_util.h"
#include "ofdm_txrx_mimo_regMacros.h"

//Only un-comment if you have a WARP user I/O board in slot 1
#define USE_USERIO_BOARD

Macframe templatePkt;

///Place holder for TRANSMIT or RECEIVE state
unsigned char trafficGen_mode;

///Measures time for beginning and end of transmission
XTime timestamp_start, timestamp_end;

unsigned int pktCount_good, pktCount_bad;

///Global Structs
warpnetEthernetPktHeader txEthPktHeader;
warpnodeID myNodeID;
warpnodeCommand myNodeCmd;

///Global Stat struct
warpnodeStats myStats[WARPNET_NUMNODES];

///Parameters used for warpmac_startPacketGeneration
unsigned int trafficGen_interval;
unsigned short trafficGen_pktLen;

unsigned int autoResp_matchCond;
unsigned int autoResp_action;
unsigned char pktBuf_tx_ACK;
unsigned char pktBuf_tx_DATA;
unsigned char pktBuf_rx;

//Global variables to for user I/O board outputs
unsigned char userIOBoard_LEDs;
unsigned char charsToPrint[16];
unsigned int pktCount_good, pktCount_bad;

//Arrays to track pkt sequence numbers for each partner node
unsigned char rxSequences[16];
unsigned char txSequences[16];

unsigned char maximumReSend;

///MAC address table with agreed upon mapping between dipswitches and physical addresses
typedef struct {
	unsigned char addr[6];
} nodeAddr;


///Byte array containing physical MAC address of this node
unsigned char myAddr[6];
///Index to the routing table that identifies this node
unsigned short int myID;

///Full rate modulation selection; QPSK by default
unsigned int pktFullRate;

//Payload code rate selection; must be CODE_RATE_NONE for uncoded PHY
unsigned int pktCodeRate;

///Buffer for holding a packet-to-xmit across multiple retransmissions
Macframe txBuffer;
///Buffer to hold received packet
Macframe rxBuffer;

///Current 802.11 channel
unsigned char chan;

//Define handy macros for CSMA MAC packet types
///Data packet with payload meant for Ethernet transmission
#define DATAPACKET 1
///Acknowledgement packet meant for halting retransmissions
#define ACKPACKET 0

///@brief Callback for the depression of the left push button
///
///This function is empty by default
void left(){
}

///@brief Callback for the depression of the right push button
///
///This button switched between different fullrate modulation orders
void right(){
	switch(pktFullRate){
		case HDR_FULLRATE_BPSK:
			pktFullRate = HDR_FULLRATE_QPSK;
			xil_printf("QPSK\r\n");
			break;
		case HDR_FULLRATE_QPSK:
			pktFullRate = HDR_FULLRATE_QAM_16;
			xil_printf("16-QAM\r\n");
			break;
		case HDR_FULLRATE_QAM_16:
			pktFullRate = HDR_FULLRATE_QAM_64;
			xil_printf("64-QAM\r\n");
			break;
		case HDR_FULLRATE_QAM_64:
			pktFullRate = HDR_FULLRATE_BPSK;
			xil_printf("BPSK\r\n");
			break;
	}
}

///@brief Callback for the depression of the up push button
///
///This button increments the 2.4GHz channel being used; only valid channels (in [1,14]) will be used
void up(){
	unsigned int newFreq;
	
	chan = (chan > 13) ? 14 : chan+1;
	newFreq = warpphy_setChannel(GHZ_2, chan);
	warpmac_leftHex(chan);
	
	xil_printf("New Frequency %d\r\n", newFreq);
}

///@brief Callback for the depression of the middle push button
///
///This button decrements the 2.4GHz channel being used; only valid channels (in [1,14]) will be used
void middle(){
	unsigned int newFreq;
	
	chan = (chan < 2) ? 1 : chan-1;
	newFreq = warpphy_setChannel(GHZ_2, chan);
	warpmac_leftHex(chan);
	
	xil_printf("New Frequency %d\r\n", newFreq);
}


///@brief Updates the LCD
///
///Updates the LCD. This is only for the User I/O daughtercard. If you do not have this card,
///this function will do nothing. This operation takes some time, so it can only be called in a few places so as to
///not adversely affect the performance of the MAC.
void updateLCD()
{
#ifdef USE_USERIO_BOARD
	warp_userioboard_set_leds(userIOBoard_LEDs);
	
	snprintf(charsToPrint, 16, "Good: %09d ", pktCount_good);
	warp_userio_lcd_printline(charsToPrint, 16, 4, 1);
	
	snprintf(charsToPrint, 16, "Bad:  %09d ", (pktCount_bad)>>1);//bad pkts get counted twice
	warp_userio_lcd_printline(charsToPrint, 16, 5, 1);
#endif	
	return;
}

///@brief Callback for the reception of UART bytes
///@param uartByte ASCII byte received from UART
///
///Provides the user with the bytes that was received over the serial port. This is useful for configuring
///PHY and MAC parameters in real time on a board.
void uartRecv_callback(unsigned char uartByte)
{
	if(uartByte != 0x0)
	{
		xil_printf("(%c)\t", uartByte);
		
		switch(uartByte)
		{
			case ASCII_1:
				pktFullRate = HDR_FULLRATE_BPSK;
				xil_printf("Tx Full Rate = BPSK\r\n");
				break;
				
			case ASCII_2:
				pktFullRate = HDR_FULLRATE_QPSK;
				xil_printf("Tx Full Rate = QPSK\r\n");
				break;
				
			case ASCII_4:
				pktFullRate = HDR_FULLRATE_QAM_16;
				xil_printf("Tx Full Rate = 16-QAM\r\n");
				break;
				
			case ASCII_6:
				pktFullRate = HDR_FULLRATE_QAM_64;
				xil_printf("Tx Full Rate = 64-QAM\r\n");
				break;
			case ASCII_D:
				warpphy_setPktDetMinDurPlus(1);
				break;
			case ASCII_d:
				warpphy_setPktDetMinDurMinus(1);
				break;
			case ASCII_P:
				warpphy_setPktDetPlus(100);
				break;
			case ASCII_p:
				warpphy_setPktDetMinus(100);
				break;
			case ASCII_F:
				if(chan<14) chan++;
				warpphy_setChannel(GHZ_2, chan);
				xil_printf("Current channel: %d\r\n",chan);
				break;
			case ASCII_f:
				if(chan>1) chan--;
				warpphy_setChannel(GHZ_2, chan);
				xil_printf("Current channel: %d\r\n",chan);
				break;
			case ASCII_C:
				warpphy_setCSMAPlus(100);
				break;
			case ASCII_c:
				warpphy_setCSMAMinus(100);
				break;
			case ASCII_S:
				warpphy_setAntennaMode(TX_ANTMODE_SISO_ANTB, RX_ANTMODE_SISO_ANTB);
				xil_printf("SISO Mode - Radio 3 Selected\r\n");
				break;
			case ASCII_s:
				warpphy_setAntennaMode(TX_ANTMODE_SISO_ANTA, RX_ANTMODE_SISO_ANTA);
				xil_printf("SISO Mode - Radio 2 Selected\r\n");
				break;
			case ASCII_a:
			case ASCII_A:
				warpphy_setAntennaMode(TX_ANTMODE_ALAMOUTI_2ANT, RX_ANTMODE_ALAMOUTI_SELDIV);
				xil_printf("Alamouti Mode - 2-antenna Tx - Selection Diversity Rx\r\n");
				break;
			case ASCII_G:
				warpphy_setTargetPlus();
				break;
			case ASCII_g:
				warpphy_setTargetMinus();
				break;
			case ASCII_X:
				warpphy_incrementTxScaling(10, 0);
				break;
			case ASCII_x:
				warpphy_incrementTxScaling(-10, 0);
				break;
			case ASCII_Y:
				warpphy_incrementTxScaling(0, 10);
				break;
			case ASCII_y:
				warpphy_incrementTxScaling(0, -10);
				break;
				
			default:
				xil_printf("Undefined command\r\n");
				break;
		}
	}
	
	return;
}
///@brief Callback for the expiration of timers
///
///This function is responsible for handling #TIMEOUT and #BACKOFF.
///The job responsibilities of this function are to:
///-increase the contention window upon the expiration of a #TIMEOUT
///-initiate a #BACKOFF timer upon the expiration of a #TIMEOUT
///-retransmit a packet upon the expiration of a #BACKOFF
///@param timerType #TIMEOUT or #BACKOFF
void timer_callback(unsigned char timerType){
	int status;
	switch(timerType){
		case TIMEOUT_TIMER:
			warpmac_setTimer(BACKOFF_TIMER);
			break;
		case BACKOFF_TIMER:
			if(txBuffer.remainingTx){
				//Copy the header over to packet buffer 1
				warpmac_prepPhyForXmit(&txBuffer, pktBuf_tx_DATA);
				//warpmac_setDebugGPIO(0xF);
				//Send packet buffer 1
				warpmac_startPhyXmit(pktBuf_tx_DATA);
				//Wait for it to finish and enable the receiver
				warpmac_finishPhyXmit();
				//warpmac_setDebugGPIO(0);
				//Start a timeout timer
				//xil_printf("----SET TIMEOUT\r\n");
				warpmac_setTimer(TIMEOUT_TIMER);
				warpmac_decrementRemainingReSend(&txBuffer);
				//Copy the header over to packet buffer 1
				//warpmac_prepPhyForXmit(&txBuffer,1);
			}
			else warpmac_enableDataFromNetwork();
			break;
	}
}


///@brief Callback for the reception of Ethernet packets
///
///This function is called by the ethernet MAC drivers
///when a packet is available to send. This function fills
///the Macframe transmit buffer with the packet and sends
///it over the OFDM link
///@param length Length, in bytes, of received Ethernet frame
void dataFromNetworkLayer_callback(Xuint32 length, char* payload){
	unsigned char rxSeqNum;
	unsigned char destNode;
	
	warpmac_resetCurrentCW();
	warpmac_disableDataFromNetwork();
	txBuffer.header.length = length;
	txBuffer.header.pktType = DATAPACKET;
	
	//Set the modulation scheme for the packet's full-rate symbols
	txBuffer.header.fullRate = pktFullRate;
	
	//Set the code rate for the packet's payload
	txBuffer.header.codeRate = pktCodeRate;
	
	//For now, assume our destination is our opposite ID (all traffic is 0 <-> 1)
	destNode = (myID+1)%2;
	
	//Use the next sequence number for this node (top four bits) and resend count of 0 (bottom four bits)
	txSequences[destNode] = (txSequences[destNode] + 1) % 256;
	txBuffer.remainingTx = (maximumReSend+1);
	txBuffer.header.seqNum = txSequences[destNode];
	
	//Copy in the packet's destination MAC address
	txBuffer.header.destAddr = (unsigned short int)(NODEID_TO_ADDR(destNode));
	
	if(warpmac_carrierSense())
	{
		//If the modium is idle:

		//Copy the header over to packet buffer 1
		warpmac_prepPhyForXmit(&txBuffer, pktBuf_tx_DATA);

		//Transmit the packet
		warpmac_startPhyXmit(pktBuf_tx_DATA);

		//Wait for it to finish
		warpmac_finishPhyXmit();

		//Start a timeout timer
		warpmac_setTimer(TIMEOUT_TIMER);
		warpmac_decrementRemainingReSend(&txBuffer);
	}
	else
	{
		//Medium was busy; start a backoff timer
		warpmac_setTimer(BACKOFF_TIMER);
	}
	
	return;
}

void sendMgmtAck(char rxSeqNum)
{
#if PrintDebug >0
	xil_printf("sent ack packet w sequence number: %d\r\n",rxSeqNum);
#endif
	int pktBuf_payloadAddr;
	
	//Fill in a node command reply pkt
	myNodeCmd.structType = STRUCTID_NODECOMMAND;
	myNodeCmd.nodeID = myID;
	myNodeCmd.cmdID = NODECMD_NODEACK;
	myNodeCmd.cmdParam = rxSeqNum;
	
	//Fill in the ethernet packet's header info
	txEthPktHeader.ethType = WARPNET_ETHTYPE_NODE2SVR;
	txEthPktHeader.srcAddr[0]=0x00;
	txEthPktHeader.srcAddr[1]=0x50;
	txEthPktHeader.srcAddr[2]=0xC2;
	txEthPktHeader.srcAddr[3]=0x63;
	txEthPktHeader.srcAddr[4]=0x3F;
	txEthPktHeader.srcAddr[5]=0x80+myID;
	
	
	memset(&(txEthPktHeader.dstAddr), (unsigned char)0xFF, 6); //Broadcast destination address
	
	txEthPktHeader.pktLength = sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeCommand);
	txEthPktHeader.numStructs = 1;
	txEthPktHeader.seqNum = rxSeqNum;
	
	pktBuf_payloadAddr = warpphy_getBuffAddr(WARPNET_NODE2SVR_PKTBUFFINDEX);
	memcpy((void *)pktBuf_payloadAddr, &(txEthPktHeader), sizeof(warpnetEthernetPktHeader));
	memcpy((void *)pktBuf_payloadAddr+sizeof(warpnetEthernetPktHeader), &(myNodeCmd), sizeof(warpnodeCommand));
	
	//Send the packet over the wire	
	warpmac_sendRawEthernetPacket((void *)pktBuf_payloadAddr, txEthPktHeader.pktLength);
	return;
}

///@brief Send the current Stat Packet over the Ethernet
///
///This function is when certain node commands request stat packets to be send over the ethernet.
//The stat package ethernet header increases the seqNum sent over wire
void sendStatsPacket()
{
	//Send a new stats packet over the wire
	
	unsigned int pktBuf_payloadAddr;
	
	//Fill in the ethernet packet's header info
	txEthPktHeader.pktLength = sizeof(warpnetEthernetPktHeader) + (WARPNET_NUMNODES*sizeof(warpnodeStats));
	txEthPktHeader.numStructs = WARPNET_NUMNODES;
	txEthPktHeader.seqNum++;
	//rest of txEthPktHeader was populated in main()
	
	//Copy the ethernet header and stats payload to a memroy buffer
	pktBuf_payloadAddr = warpphy_getBuffAddr(WARPNET_NODE2SVR_PKTBUFFINDEX);
	memcpy((void *)pktBuf_payloadAddr, &(txEthPktHeader), sizeof(warpnetEthernetPktHeader));
	
	//Copy each stats struct (one per partner)
	memcpy((void *)pktBuf_payloadAddr+sizeof(warpnetEthernetPktHeader), &myStats, WARPNET_NUMNODES*sizeof(warpnodeStats));
	
	//Send the packet over the wire
	warpmac_sendRawEthernetPacket((void *)pktBuf_payloadAddr, txEthPktHeader.pktLength);
	
	return;
}

///@brief Update node parameters with warpnodeControl Struct values
///
///This function process the warpnodeControl struct by taking the specific
///paramters and modifying the current board configuration
///@param packet Pointer to warpnodeControl
void processControlStruct(warpnodeControl* ctrlStruct)
{

	//Interpret the control struct's modulation rate value
	switch(ctrlStruct->modOrder)
	{
		case 0x1: //BPSK
			pktFullRate = HDR_FULLRATE_BPSK;
			warp_userio_lcd_printline("Mod order: BPSK ", 16, 8, 1);
			break;
		case 0x2: //QPSK
			pktFullRate = HDR_FULLRATE_QPSK;
			warp_userio_lcd_printline("Mod order: QPSK ", 16, 8, 1);
			break;
		case 0x4: //16-QAM
			pktFullRate = HDR_FULLRATE_QAM_16;
			warp_userio_lcd_printline("Mod order: 16QAM", 16, 8, 1);
			break;
		case 0x6: //64-QAM
			pktFullRate = HDR_FULLRATE_QAM_64;
			warp_userio_lcd_printline("Mod order: 64QAM", 16, 8, 1);
			break;
			/*******************/
		default:
			//Invalid value, so ignore it and leave txFulrate as-is
			xil_printf("Invalid Mod order - Recieved: %x\r\n", ctrlStruct->modOrder);
			break;
	}
	
	//Set the transmit power for both radios
	warpphy_setTxPower(ctrlStruct->txPower);
	
	snprintf(charsToPrint, 16, "Tx Pwr: 0x%2x    ", (ctrlStruct->txPower));
	warp_userio_lcd_printline(charsToPrint, 16, 6, 1);
	
	//Set the center frequency of both radios
	// Values 1-14 are 2.4GHz channels
	// Values 15-37 are 5GHz channels
	// Any other value is invalid and ignored
	if( ((ctrlStruct->channel) > 0) && ((ctrlStruct->channel) <= 37) )
	{
		if((ctrlStruct->channel) <= 14)
			warpphy_setChannel(GHZ_2, ctrlStruct->channel);
		else
			warpphy_setChannel(GHZ_5, (ctrlStruct->channel) - 14);
	}
	
	snprintf(charsToPrint, 16, "Channel: %2d     ", (ctrlStruct->channel));
	warp_userio_lcd_printline(charsToPrint, 16, 7, 1);
	
	return;
}

void mgmtFromNetworkLayer_callback(Xuint32 length, char* payload){
	
	//Bookkeeping variables
	int pktBuf_payloadAddr;
	unsigned char numRxStructs, thisStructType;
	unsigned int rxPktOffset;
	int i;

	unsigned char rxSeqNum=0;
	unsigned char destNode;

	warpmac_resetCurrentCW();

	//Re-interpret the Rx packet as a WARPnet Ethernet frame
	warpnetEthernetPktHeader* receivedPkt;
	
	//Pointers to each kind of possible received struct; used to re-interpret the packet below
	warpnodeID* idStruct;
	warpnodeControl* controlStruct;
	warpnodeCommand* cmdStruct;
	warpnodeTraffic* trafficStruct;
	
	receivedPkt = (warpnetEthernetPktHeader*) payload;
	
	//Ignore node-to-server packets from other nodes
	if ((receivedPkt->ethType) == WARPNET_ETHTYPE_NODE2SVR)
	{
		return;
	}
	
	//If this is a WARPnet Ethernet packet, process it immediately
	if((receivedPkt->ethType) == WARPNET_ETHTYPE_SVR2NODE)
	{
		rxPktOffset = sizeof(warpnetEthernetPktHeader);
		numRxStructs = receivedPkt->numStructs;
		rxSeqNum = receivedPkt->seqNum;
		
		//Itereate over each struct in the Ethernet packet
		for(i=0; i<numRxStructs; i++)
		{
			//Punt if numStructs would have us march past the end of the received packet
			if(rxPktOffset >= length)
			{
				xil_printf("Error! Mgmt pktLength too short for numStructs\r\n");
				return;
			}
			
			//Extract the struct type (always the first byte of a struct)
			thisStructType = *(unsigned char*)(payload+rxPktOffset);
			switch(thisStructType)
			{
				case STRUCTID_NODEID:
					//NodeID request; node must response with a nodeID reply
					
					//Initialize my nodeID struct (to respond to server inquiries)
					myNodeID.structType = STRUCTID_NODEID;
					myNodeID.nodeID = myID;
					
					txEthPktHeader.pktLength = sizeof(warpnetEthernetPktHeader) + sizeof(warpnodeID);
					txEthPktHeader.numStructs = 1;
					
					pktBuf_payloadAddr = warpphy_getBuffAddr(WARPNET_NODE2SVR_PKTBUFFINDEX);
					memcpy((void *)pktBuf_payloadAddr, &(txEthPktHeader), sizeof(warpnetEthernetPktHeader));
					memcpy((void *)pktBuf_payloadAddr+sizeof(warpnetEthernetPktHeader), &(myNodeID), sizeof(warpnodeID));
					
					warpmac_sendRawEthernetPacket((void *)pktBuf_payloadAddr, txEthPktHeader.pktLength);
					
					rxPktOffset += sizeof(warpnodeID);
					break;
					
				case STRUCTID_NODECONTROL:
					//Node control packet; update active parameters and respond with an ACK if requested
					controlStruct = (warpnodeControl*)(payload+rxPktOffset);
					
					if( (controlStruct->nodeID) == myID)
					{
						processControlStruct((warpnodeControl*)(payload+rxPktOffset));
						
						//Send Ack Packet if the server used a non-zero sequence number in its packet
						if(rxSeqNum > 0) sendMgmtAck(rxSeqNum);
					}
					
					rxPktOffset += sizeof(warpnodeControl);
					break;
					
				case STRUCTID_NODECOMMAND:
					//Node command struct; action depends on second parameter (cmdID)
					cmdStruct = (warpnodeCommand*)(payload+rxPktOffset);
					
					switch(cmdStruct->cmdID)
					{
						case NODECMD_RESETSTATS:
							//Reset all locally stored statistics
							if( (cmdStruct->nodeID) == myID)
							{	
								for(i=0; i<WARPNET_NUMNODES; i++)
								{
									myStats[i].goodPackets = 0;
									myStats[i].partnerBadPackets = 0;
									myStats[i].otherBadPackets = 0;
									myStats[i].txBytes = 0;
									myStats[i].rxBytes = 0;
									myStats[i].time = 0;
								}
								
								//Send Ack Packet if the server used a non-zero sequence number in its packet
								if(rxSeqNum > 0) sendMgmtAck(rxSeqNum);
							}
							break;
							
						case NODECMD_START:
							//Start an experiment

							//Record the current time
							XTime_GetTime(&timestamp_start);
						
							//Start generating traffic (Mode is set by a previous warpnodeTraffic)
							if (trafficGen_mode == TRANSMIT)
							{
								warpmac_startPacketGeneration(trafficGen_pktLen, trafficGen_interval);
							}
						
						break;
						
						case NODECMD_STOP:
							//Stop an experiment
						
							//Stop generating traffic (ok to call this even if this node wasn't generating traffic)	
							warpmac_stopPacketGeneration();
							
							//Record the time and calculate the experiment duration
							XTime_GetTime(&timestamp_end);
							long long xtime=0;
							xtime = timestamp_end - timestamp_start;
							
							//Store the duration in each stats struct (yes, it's redundant; but this way each stats struct is meaningful alone)
							for(i=0; i<WARPNET_NUMNODES; i++)
							{
								myStats[i].time = xtime;
							}
							
							if((cmdStruct->nodeID) == myID)
							{	
								//Send Ack Packet if the server used a non-zero sequence number in its packet
								if(rxSeqNum > 0) sendMgmtAck(rxSeqNum);
							}
							
							break;
							
						case NODECMD_REQUESTSTATS:
							
							if( (cmdStruct->nodeID) == myID)
							{	
								sendStatsPacket();
							}
							break;
						
						default:
						
							xil_printf("Invalid node command: 0x%x\r\n", cmdStruct->cmdID);
						break;
					}/*END switch(cmdStruct->cmdID)*/
					
					rxPktOffset += sizeof(warpnodeCommand);
					break;
					
				case STRUCTID_NODETRAFFIC:
					//Node traffic struct; configures traffic generation options

					trafficStruct = (warpnodeTraffic*)(payload+rxPktOffset);
					
					if((trafficStruct->nodeID) == myID)
					{
						//Update the traffic generation parameters
						trafficGen_interval	= (trafficStruct->txInterval);
						trafficGen_mode		= (trafficStruct->trafficMode);
						trafficGen_pktLen	= (trafficStruct->txPacketLen);

						//Send Ack Packet if the server used a non-zero sequence number in its packet
						if(rxSeqNum > 0) sendMgmtAck(rxSeqNum);
					}
					
					rxPktOffset += sizeof(warpnodeTraffic);

					break;
					
				default:
					//Didn't understand the WARPnet packet; discard it
					xil_printf("Got unknown management pkt type: 0x%x\r\n", thisStructType);
					break;
			}/* END switch(thisStructType) */
			
		}/* END for(numRxStructs) */
	}/* END if(warpNet packet) */
	
	return;
}

///@brief Callback for the reception of bad wireless headers
///
///@param packet Pointer to received Macframe
void phyRx_badHeader_callback() {
	warpmac_incrementLEDLow();
	
	userIOBoard_LEDs = userIOBoard_LEDs==0x80 ? 0x01 : (userIOBoard_LEDs << 1);

	pktCount_bad++;
	return;
}

///@brief Callback for the reception of good wireless headers
///
///This function then polls the PHY to determine if the entire packet passes checksum
///thereby triggering the transmission of the ACK and the transmission of the received
///data over Ethernet.
///@param packet Pointer to received Macframe
void phyRx_goodHeader_callback(Macframe* packet){
	
	unsigned char state = PHYRXSTATUS_INCOMPLETE;
	unsigned char srcNode;
	unsigned char shouldSend = 0;
	int i;
	
	//Calculate the node ID from the packet's source MAC address
	srcNode = ADDR_TO_NODEID( (packet->header.srcAddr) );
	
	//Rotate the ring of LEDs clockwise
	userIOBoard_LEDs = userIOBoard_LEDs==0x1 ? 0x80 : (userIOBoard_LEDs >> 1);
	
	//Print the header (debug only- completely ruins actual performance)
#if 0
	for(i=0; i<24; i++)
	{
		xil_printf("%d.",*((unsigned char*)&(packet->header) + i));
	}
	xil_printf("\r\n");
#endif
	
	//If the packet is addressed to this node
	if( packet->header.destAddr == (NODEID_TO_ADDR(myID)) )
	{
		switch(packet->header.pktType){
			//If received packet is data
			case DATAPACKET:
				//At this point, we have pre-loaded the PHY transmitter with the ACK in hoping that
				//the packet passes checksum. Now we wait for the state of the received to packet
				//to move from PHYRXSTATUS_INCOMPLETE to either PHYRXSTATUS_GOOD or PHYRXSTATUS_BAD

				//If it's a long data packet (i.e. we have a while to wait), update the LCD with some stats
				if((packet->header.length)>1400){
					updateLCD();
				}
				
				//Blocks until the PHY declares the payload good or bad
				state = warpmac_finishPhyRecv();
				
				if(state & PHYRXSTATUS_GOOD){
					//The auto-reponder will send the pre-programmed ACK automatically
					//User code only needs to update its stats, then check to see the PHY
					// is finished transmitting

					//Toggle the top LEDs
					warpmac_incrementLEDHigh();
					
					//Update the right-hex display with the current sequence number
					warpmac_leftHex(0xF & (packet->header.seqNum));
					
					//Check if the last received seq number for this partner node matches this received pkt
					// If not, record the new number and allow the pkt to be forwarded over the wire
					if(srcNode<WARPNET_NUMNODES && rxSequences[srcNode] != (packet->header.seqNum))
					{
						//Not a duplicate packet; update this partner's last-known sequence number
						rxSequences[srcNode] = (packet->header.seqNum);
						shouldSend = 1;
					} 
					
					//Starts the DMA transfer of the payload into the EMAC
					if(shouldSend) warpmac_prepDataToNetwork(packet);
					
					//Blocks until the PHY is finished sending and enables the receiver
					warpmac_finishPhyXmit();
					
					//Waits until the DMA transfer is complete, then starts the EMAC
					if(shouldSend) warpmac_startDataToNetwork(packet);
					
					pktCount_good++;
				}
				
				if(state & PHYRXSTATUS_BAD){
					warpmac_incrementLEDLow();
					pktCount_bad++;
					
					//Rotate the ring of LEDs counter-clockwise
					userIOBoard_LEDs = userIOBoard_LEDs==0x80 ? 0x01 : (userIOBoard_LEDs << 1);
				}
				
				break;
				
			case ACKPACKET:
				//Clear the TIMEOUT and enable Ethernet
				if(warpmac_inTimeout()){
					warpmac_incrementLEDHigh();

					//Clear the timeout timer, set when we transmitted the data packet
					warpmac_clearTimer(TIMEOUT_TIMER);

					//Clear the remaining transmit count to assure this packet won't be re-transmitted
					txBuffer.remainingTx = 0;
					
					//Start a backoff, to gaurantee a random time before attempting to transmit again
					warpmac_setTimer(BACKOFF_TIMER);
				}
				
				pktCount_good++;
				
				break;
		}
	}
}

///@brief Main function
///
///This function configures MAC parameters, enables the underlying frameworks, and then loops forever.
int main(){		
	int i;
	
	//Initialize global variables
	chan = 9;
	
	//Assign the packet buffers in the PHY
	// The auto responder can't transmit from buffer 0, so we use it for Rx packets
	// The other assignments (DATA/ACK) are arbitrary; any buffer in [1,30] will work
	pktBuf_rx = 0;
	pktBuf_tx_DATA = 1;
	pktBuf_tx_ACK = 2;

	//Set the full-rate modulation to QPSK by default
	pktFullRate = HDR_FULLRATE_QPSK;
	
	//Set the payload coding rate to NONE by default
	//Note: Reference Design 13 PHY does not include channel coding
	pktCodeRate = CODE_RATE_NONE;
	
	//Initialize the MAC/PHY frameworks
	warpmac_init();
	maximumReSend = 8;
	warpmac_setMaxResend(maximumReSend);
	warpmac_setMaxCW(5);
	warpmac_setTimeout(100);
	warpmac_setSlotTime(22);

	//Read Dip Switch value from FPGA board.
	//This value will be used as an index into the routing table for other nodes
	myID = (unsigned short int)warpmac_getMyId();

	//Configure the PHY and radios for single antenna (SISO) mode
	warpphy_setAntennaMode(TX_ANTMODE_SISO_ANTA, RX_ANTMODE_SISO_ANTA);
	//warpphy_setAntennaMode(TX_ANTMODE_ALAMOUTI_ANTA, RX_ANTMODE_ALAMOUTI_ANTA);
			
	//Rx buffer is where the EMAC will DMA Wireless payloads from
	warpmac_setRxBuffer(&rxBuffer, pktBuf_rx);
	//Tx buffer is where the EMAC will DMA Ethernet payloads to
	warpmac_setTxBuffer(pktBuf_tx_DATA);
	
	//Set the modulation scheme use for base rate (header) symbols
	warpmac_setBaseRate(QPSK);

	//Copy this node's MAC address into the Tx buffer's source address field
	txBuffer.header.srcAddr = (unsigned short int)(NODEID_TO_ADDR(myID));
	
	//Register callbacks
	warpmac_setBadHeaderCallback((void *)phyRx_badHeader_callback);
	warpmac_setGoodHeaderCallback((void *)phyRx_goodHeader_callback);
	warpmac_setTimerCallback((void *)timer_callback);
	warpmac_setDataFromNetworkCallback((void *)dataFromNetworkLayer_callback);
	warpmac_setMgmtFromNetworkCallback((void *)mgmtFromNetworkLayer_callback);
	warpmac_setUpButtonCallback((void *)up);
	warpmac_setMiddleButtonCallback((void *)middle);
	warpmac_setLeftButtonCallback((void *)left);
	warpmac_setRightButtonCallback((void *)right);
	warpmac_setUartRecvCallback((void *)uartRecv_callback);

	//Set the default center frequency
	warpphy_setChannel(GHZ_2, chan);
	
	//Enable carrier sensing
	warpmac_enableCSMA();
	
	/*** WARPnet Measurement/Control Setup ***/
	//Construct an Ethernet pkt header to use when sending packets to the server
	//srcAddr is arbitrary; dstAddr is always the broadcast address
	txEthPktHeader.ethType = WARPNET_ETHTYPE_NODE2SVR;
	txEthPktHeader.srcAddr[0]=0x00;
	txEthPktHeader.srcAddr[1]=0x50;
	txEthPktHeader.srcAddr[2]=0xC2;
	txEthPktHeader.srcAddr[3]=0x63;
	txEthPktHeader.srcAddr[4]=0x3F;
	txEthPktHeader.srcAddr[5]=0x80+myID;
	memset(&(txEthPktHeader.dstAddr), (unsigned char)0xFF, 6); //Broadcast destination address
	
	//Initialize the stats structs and sequence number vectors
	for(i=0; i<WARPNET_NUMNODES; i++)
	{
		//Sequence numbers (to identify received duplicate packets)
		txSequences[i] = 1;
		rxSequences[i] = 0;

		//Stats structs, one per node in the network
		myStats[i].structType = STRUCTID_NODESTATS;
		myStats[i].nodeID = myID;
		myStats[i].partnerID = i;
		myStats[i].goodPackets = 0;
		myStats[i].otherBadPackets = 0;
		myStats[i].partnerBadPackets = 0;
		myStats[i].rxBytes = 0;
		myStats[i].txBytes = 0;
		myStats[i].reserved0 = 0;
	}
	
	/******** START autoResponse setup *******/
	//Setup the PHY's autoResponse system
	// For CSMA, it is configured to send pktBuf pktBuf_tx_ACK when a good DATA packet is received addressed to this node
	
	//Match condition 0: received header's destination address is this node's address
	autoResp_matchCond = PHY_AUTORESPONSE_MATCH_CONFIG(PKTHEADER_INDX_DSTADDR, 2, htons(NODEID_TO_ADDR(myID)));
	mimo_ofdmTxRx_setMatch0(autoResp_matchCond);

	//Match condition 1: received header's type is DATA 
	autoResp_matchCond = PHY_AUTORESPONSE_MATCH_CONFIG(PKTHEADER_INDX_TYPE, 1, DATAPACKET);
	mimo_ofdmTxRx_setMatch1(autoResp_matchCond);
	
	//Configure the header translator to use the Rx pkt's src address as the outgoing pkt's dst address
	// Addresses are two bytes, so two entries in the header translator need to be overridden
	// Except for these bytes, the ACK pktBuf's contents will be sent unaltered
	// PHY_HEADERTRANSLATE_SET(templatePktBuf, byteAddrToOverwrite, srcPktBuf, srcByteAddr)
	PHY_HEADERTRANSLATE_SET(pktBuf_tx_ACK, (PKTHEADER_INDX_DSTADDR+0), pktBuf_rx, (PKTHEADER_INDX_SRCADDR+0));
	PHY_HEADERTRANSLATE_SET(pktBuf_tx_ACK, (PKTHEADER_INDX_DSTADDR+1), pktBuf_rx, (PKTHEADER_INDX_SRCADDR+1));

	//Create a template ACK packet
	templatePkt.header.fullRate = pktFullRate;
	templatePkt.header.codeRate = pktCodeRate;
	templatePkt.header.length = 0;
	templatePkt.header.srcAddr = (unsigned short)(NODEID_TO_ADDR(myID));
	templatePkt.header.pktType = ACKPACKET;
	
	//Copy the header down to the PHY's packet buffer
	// This doesn't actually send anything; the autoResponse system will use this template when sending ACKs
	warpmac_prepPhyForXmit(&templatePkt, pktBuf_tx_ACK);

	//Action defitions come last; bad things might happen if an action is enabled (set non-zero) before the template pkt is ready.
	//All actors are disabled during warpphy_init; only load non-zero configurations for actors you intend to use

	//Action 0: send pkt from buf pktBuf_tx_ACK when match0 & match1 & goodPkt, using header translation
	autoResp_action = PHY_AUTORESPONSE_TXACTION_CONFIG(pktBuf_tx_ACK, PHY_AUTORESPONSE_ACT_TRANS_HDR, 0, (PHY_AUTORESPONSE_REQ_MATCH0 | PHY_AUTORESPONSE_REQ_MATCH1 | PHY_AUTORESPONSE_REQ_GOODHDR | PHY_AUTORESPONSE_REQ_GOODPKT));
	mimo_ofdmTxRx_setAction0(autoResp_action);
	/******* END autoResponse setup ******/
	
	//Listen for new packets to send (either from Ethernet or local dummy packets)
	warpmac_enableDataFromNetwork();
	
#ifdef USE_USERIO_BOARD
	userIOBoard_LEDs = 1;
	//Initialize the LCD for slot 1 (i.e. don't flip the image) and non-inverted colors
	warp_userioboard_lcd_init(1, 0);
	warp_userioboard_set_lcd_charbuf(1);
	
	//Print some static lines to the user I/O board LCD
	//snprintf(charsToPrint, 16, " Ref Design v12 ", pktCount_bad);
	warp_userio_lcd_printline("RefDesign v13.01", 16, 1, 1);
	//snprintf(charsToPrint, 16, "  Rx Pkt Counts", pktCount_bad);
	warp_userio_lcd_printline("  Rx Pkt Counts ", 16, 3, 1);
	
	//Set the user I/O board LEDs
	warp_userioboard_set_leds(userIOBoard_LEDs);
#endif
	
	xil_printf("Reference Design v13.01 CSMAMAC\r\n");
	xil_printf("Beginning main loop\r\n");
	
	while(1)
	{
		//Poll the timer, PHY and user I/O forever; actual processing will happen via callbacks above
		warpmac_pollPeripherals();
	}
	
	return;
}
